import pandas as pd


def _backward_fill(X):
    return X.fillna(method="bfill").fillna(method="ffill")

def _forward_fill(X):
    return X.fillna(method="ffill").fillna(method="bfill")

def _mean_fill(X):
    if isinstance(X, pd.DataFrame):
        for i in X.columns[X.isnull().any(axis=0)]:
            X[i].fillna(X[i].mean(), inplace=True)
    else:
        X.fillna(X.mean(), inplace=True)
    return X

def _median_fill(X):
    if isinstance(X, pd.DataFrame):
        for i in X.columns[X.isnull().any(axis=0)]:
            X[i].fillna(X[i].median(), inplace=True)
    else:
        X.fillna(X.median(), inplace=True)
    return X


from reeco_ml_preprocessing.base import BasePreprocessor


class DataSlicer(BasePreprocessor):

    def __init__(self, start_index: int = None, end_index: int = None):
        self.start_index = start_index
        self.end_index = end_index

    def transform(self, X):
        if self.start_index is None and self.end_index is None:
            return X
        elif self.start_index is None:
            return X[:self.end_index]
        elif self.end_index is None:
            return X[self.start_index:]
        else:
            return X[self.start_index:self.end_index]

    def fit_transform(self, X):
        return self.transform(X)


from lunarcalendar import Converter, Solar
import numpy as np
import pandas as pd

class LunarConcater:

    def __init__(
        self,
        month: bool = True,
        day: bool = True,
        hour: bool = True,
        year: bool = False,
        minute: bool = False,
        second: bool = False
    ):
        self.mask = [year, month, day, hour, second, minute]
        # Note that all title must has prefix "static_" in order to differentiate between
        # static covariates and others
        self.title = [
            'static_lunar_year',
            'static_lunar_month',
            'static_lunar_day',
            'static_lunar_hour',
            'static_lunar_second',
            'static_lunar_minute'
        ]

    def transform(self, X: pd.DataFrame):
        n_components = len(self.mask)
        lunar_info = X.index.to_pydatetime()
        lunar_info = np.array([self.get_lunar_components(x) for x in lunar_info])
        lunar_col_kept = [self.title[i] for i in range(n_components) if self.mask[i]]
        
        df_lunar = pd.DataFrame(
            lunar_info[:, self.mask],
            columns=lunar_col_kept,
            index=X.index
        )
        return pd.concat([X, df_lunar], axis=1)
    
    def fit_transform(self, X: pd.DataFrame):
        return self.transform(X)

    def get_lunar_components(self, x: pd.DatetimeIndex):
        """Convert a Solar date into Lunar date"""
        lunar = Converter.Solar2Lunar(Solar(x.year, x.month, x.day))
        res = [lunar.year, lunar.month, lunar.day, x.hour, x.second, x.minute]
        return res



import pandas as pd
from reeco_ml_preprocessing.base import BasePreprocessor

from reeco_ml_preprocessing.time_series.simple_imputer import _mean_fill, _median_fill, _backward_fill, _forward_fill


class TimeSeriesImputer(BasePreprocessor):

    def __init__(self, method):
        """
        Method initialization for imputer.

        Args:
        -------
        method: String, default='linear'
            Custom kind to fill missing value.
            'linear': Use linear interpolation within each column.
            'backward': Use backward fill. Apply forward fill with remaining missing values in last columns.
            'forward': Use forward fill. Apply backward fill with remaining missing values in first columns.
            'mean': Replace by mean.
            'median': Replace by median.
        """
        self.method = method
        self.columns_ = None
        self.indices_ = None

    def fit(self):
        if self.method not in ["forward", "backward", "median", "mean", "linear"]:
            raise ValueError("Method {} is not exists".format(self.method))
        return self

    def fit_transform(self, X: pd.DataFrame):
        """
        Fill missing value with custom technique.

        Args:
        -------
        X: DataFrame
            Samples.
        save_columns_and_indices: Boolean
            When using fit, save the columns and indices for inverse transform later.
        create_date_col: Boolean
            When using fit, add an Time column with type float, which is retrieved from
            the indices.

        Returns:
        -------
        X_new: DataFrame
            Samples with filled values.
        """
        self.fit()
        X_new = self._impute(X, self.method)
        self.columns_ = X.columns
        self.indices_ = X.index
        return X_new

    def transform(self, X: pd.DataFrame):
        self.indices_ = X.index
        X_new = self._impute(X, self.method)
        return X_new

    def _impute(self, X: pd.DataFrame, method: str):
        if method == "forward":
            X = _forward_fill(X)
        elif method == "backward":
            X = _backward_fill(X)
        elif method == "median":
            X = _median_fill(X)
        elif method == "mean":
            X = _mean_fill(X)
        elif method == "linear":
            X = self._linear_fill(X)
        else:
            raise Exception("Cannot found method {}".format(method))
        return X.fillna(0.0)

    def _linear_fill(self, X):
        return _backward_fill(X.interpolate(method='linear'))



import numpy as np
import pandas as pd
from reeco_ml_preprocessing.base import BasePreprocessor

class PandasNumpyTransformer(BasePreprocessor):
    """
    Transform a Dataframe into Numpy ndarray while keeping the column and
    indices for inverse transformation.
    """

    def fit_transform(self, X: pd.DataFrame) -> np.ndarray:
        self.columns_ = X.columns
        self.indices_ = X.index
        return X.values

    def transform(self, X: pd.DataFrame) -> np.ndarray:
        self.indices_ = X.index
        return X.values

    def inverse_transform(self, X) -> pd.DataFrame:
        if not isinstance(X, pd.DataFrame) and not isinstance(X, pd.Series):
            return pd.DataFrame(X, index=self.indices_, columns=self.columns_)
        else:
            return X



from typing import Union

import pandas as pd


class TimeSeriesCleaner():
    """
    The TimeSeriesCleaner perform must-preprocessed steps for timeseries data.
    These includes:
    
    1. Set datetime column as index.
    2. Remove null rows exists in the label column.
    3. Align datetime index.
    """

    def __init__(
        self,
        date_col: Union[str, pd.Index],
        label_col: Union[str, pd.Index],
        sampling_rule: str,
        time_lag: str,
        limit: int = None
    ):
        self.date_col       = date_col
        self.label_col      = label_col
        self.sampling_rule  = sampling_rule
        self.time_lag       = time_lag
        self.limit          = limit

    def remove_null_rows(self, X):
        X = X.dropna(how='all')
        return X

    def set_index(self, X, date_col) -> pd.DataFrame:
        """Resample and set index. All categorical variable will also be removed."""
        if date_col not in X.columns:
            raise KeyError("{} does not appear in your data columns".format(date_col))

        try:
            # Check if the column can be converted to datetime
            X[date_col] = pd.to_datetime(X[date_col])
        except:
            raise ValueError("The expected date time column does not have the correct format. \
                Try to choose different column or change its format.")
        X = X.set_index(date_col).sort_index()
        return X

    def align_time(self, X, sampling_rule: str, time_lag: str, limit: int = None):
        """Align time to sampling_rule."""
        base = '1H' if sampling_rule.endswith('H') else sampling_rule
        index = pd.date_range(
            start=X.index[0].round(base),
            end=X.index[-1].round(base),
            freq=sampling_rule
        )
        # Fillna to 0 if all NaN within a column
        X = X.apply(lambda x: x.reindex(
            index,
            method='nearest',
            tolerance=time_lag
        ))
        if limit is not None:
            X = X[-limit:]
        return X

    def fit(self, X):
        X = self.set_index(X, self.date_col)
        X = self.remove_null_rows(X)
        X = self.align_time(X, self.sampling_rule, self.time_lag, self.limit)
        return X

    def fit_transform(self, X):
        return self.fit(X)

    def transform(self, X):
        return self.fit(X)



class BasePreprocessor:

    def __init__(self):
        pass

    def fit(self, X):
        pass

    def transform(self, X):
        raise NotImplementedError

    def fit_transform(self, X):
        self.fit(X).transform(X)

    def inverse_transfrom(self, X):
        pass


from typing import List

import joblib


class PreprocessPipeline:

    def __init__(self, pipeline):
        self.pipeline = pipeline

    def save(self, path):
        """Save the processor to expected `path` as a `.joblib` file."""
        joblib.dump(self, path)

    def fit(self):
        pass

    def transform(self, X):
        for p in self.pipeline:
            if hasattr(p, "transform"):
                X = p.transform(X)
        return X

    def fit_transform(self, X):
        for p in self.pipeline:
            if hasattr(p, "fit_transform"):
                X = p.fit_transform(X)
        return X

    def inverse_transform(self, X):
        for p in reversed(self.pipeline):
            if hasattr(p, "inverse_transform"):
                X = p.inverse_transform(X)
        return X


from typing import Union
import numpy as np
import pandas as pd
from reeco_ml_preprocessing.base import BasePreprocessor


class StandardScaler(BasePreprocessor):

    def __init__(self, target: str = None):
        self.target = target

    def fit(self, X: Union[pd.DataFrame, np.ndarray]):
        self.mean = X.mean(axis=0)
        self.std = X.std(axis=0)

    def fit_transform(self, X: Union[pd.DataFrame, np.ndarray]):
        self.fit(X)
        return (X - self.mean) / self.std

    def transform(self, X: Union[pd.DataFrame, np.ndarray]):
        if isinstance(X, pd.DataFrame):
            for col in X.columns:
                pass
        if X.ndim == 1:
            return (X - self.mean[self.target]) / self.std[self.target]
        else:
            return (X - self.mean) / self.std

    def inverse_transform(self, X: Union[pd.DataFrame, np.ndarray]):
        if X.ndim == 1:
            return X * self.std[self.target] + self.mean[self.target]
        return X * self.std + self.mean


from typing import Tuple, Union
import numpy as np
import pandas as pd

from reeco_ml_preprocessing.base import BasePreprocessor

class SlidingWindow(BasePreprocessor):
    """
    For every Series containing values `(t_0, t_1, t_2,..., t_l)`, transform
    into two 2D frame with `slides` columns. For example `(0, 1, 2, 3, 4)` when
    transforming with `slides = 2` will return:
    ```
    np.array([
        [0, 1],
        [1, 2],
        [2, 3]
    ])
    ```
    and
    np.array([[2], [3], [4]])
    The SlidingWindow will return 3D-numpy array, therefore a PandasNumpyTransformer
    will be attached within.

    Args:
    -------
    input_timesteps: int
        Input timesteps, i.e., number of columns in past dataframe
    output_timesteps: int. Depreciated
        Output timesteps, i.e., number of columns in future dataframe
        or the number of desired timesteps to be predicted.
    """

    def __init__(self, input_timesteps: int, output_timesteps: int, target: str):
        self.target = target
        self.input_timesteps = input_timesteps
        self.output_timesteps = output_timesteps

    def fit_transform(self, X: Union[pd.Series, pd.DataFrame]) -> Tuple[np.ndarray, np.ndarray]:
        if isinstance(X, pd.Series):
            rnp = self.get_historical_attributes(X).values
            return (rnp, rnp)

        elif isinstance(X, pd.DataFrame):
            n_examples = X.shape[0] - self.input_timesteps - self.output_timesteps + 1
            n_attr = len(X.columns)
            rnp = np.zeros((n_examples, n_attr, self.input_timesteps))
            tnp = np.zeros((n_examples, self.output_timesteps))
            for i, col in enumerate(X.columns):
                rnp[:, i, :] = self.get_historical_attributes(X[col])
            tnp[:] = self.get_historical_attributes(X[self.target], target=True)
            return (rnp, tnp)

    def transform(self, X: pd.Series) -> Tuple[np.ndarray, np.ndarray]:
        return self.fit_transform(X)

    def get_historical_attributes(self, series: pd.Series, target: bool = False) -> np.ndarray:
        series.fillna(0.0, inplace=True)
        df = pd.DataFrame()
        n_features = self.input_timesteps + self.output_timesteps
        
        for i in range(n_features):
            df[f'lag-{i}-{series.name}'] = series.shift(i)

        df.dropna(inplace=True)
        if not target:
            df = df.iloc[:, :self.input_timesteps].iloc[:, ::-1]
        else:
            df = df.iloc[:, -self.output_timesteps:]
        return df.values




import os
from setuptools import find_packages
from distutils.core import setup

current_directory = os.path.dirname(os.path.abspath(__file__))
try:
    with open(os.path.join(current_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
except Exception:
    long_description = ''

setup(
	name='reeco_ml_preprocessing',
	packages=find_packages('.'),
	version='0.1.4',
	description='Package for ML preprocessing',
	long_description=long_description,
	long_description_content_type='text/markdown',
	author='ARI Technology',
	author_email='dung.ut@ari.com.vn',
	url='https://gitlab.com/vn-reecotech/Ari/reeco-data-platform/preprocessor',
	download_url='https://gitlab.com/vn-reecotech/Ari/reeco-data-platform/preprocessor',
	install_requires=[
        "joblib==1.0.1",
        "LunarCalendar==0.0.9",
        "numpy",
        "pandas"
    ]
)